/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2008, 2009. All Rights Reserved.
 *  
 * U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 ******************************************************************************/

using System;
using System.Collections;
using System.Drawing;
using System.Windows.Forms;

using Rational.Test.Ft.Script;
using Rational.Test.Ft.Util;
using Rational.Test.Ft;
using Rational.Test.Ft.Sys.Graphical;
using Rational.Test.Ft.Object.Interfaces;
using Rational.Test.Ft.Sys;
using Rational.Test.Ft.Domain.Win;
using Rational.Test.Ft.Vp;
using Rational.Test.Ft.Vp.Impl;
using Rational.Test.Ft.Domain;
using Rational.Test.Ft.Domain.Net;


namespace SDK.Sample
{
	[ThreadingModel(ThreadingModel.UI_CHANNEL)]
	public class CustomDataGridProxy : ControlWithSubobjectsProxy
	{
		
		#region Constructor Declaration

		public CustomDataGridProxy(NetTestDomainImplementation domain, IChannel channel, System.Object theTestObject)
			: base(domain, channel, theTestObject)
		{
           
		}
        

		#endregion

		#region Define Role

		public override string GetRole()
		{
			return Rational.Test.Ft.Domain.TestObjectRole.ROLE_TABLE;
		}
		
		#endregion

		#region Object Map Hierarchy Methods

		public override ProxyTestObject[] GetMappableChildren()
		{
			ProxyTestObject[] mappableChildren = null;
			ProxyTestObject[] children = GetChildren();

			if (children != null)
			{
				System.Collections.ArrayList mappableChildrenArray = new System.Collections.ArrayList(children.Length);

				for (int i = 0; i < children.Length; i++)
				{
					if (children[i].ShouldBeMapped())
					{
						if (!(children[i] is Rational.Test.Ft.Domain.Net.ScrollBarProxy))
							mappableChildrenArray.Add(children[i]);
					}
					else
					{
						ProxyTestObject[] grandChildren = children[i].GetMappableChildren();
						for (int j = 0; grandChildren != null && j < grandChildren.Length; j++)
						{
							mappableChildrenArray.Add(grandChildren[j]);
						}
					}
				}
				if (mappableChildrenArray.Count > 0)
				{
					mappableChildren = new ProxyTestObject[mappableChildrenArray.Count];

					for (int i = 0; i < mappableChildrenArray.Count; i++)
					{
						mappableChildren[i] = (ProxyTestObject)mappableChildrenArray[i];
					}
				}

			}
			//Only add .classIndex if the parent is mappable.  Otherwise the incorrect
			//index could be added when the mappable parent adds the index.
			if (mappableChildren != null && ShouldBeMapped())
			{
				System.Collections.Hashtable classTable = new System.Collections.Hashtable();
				for (int iProxies = 0; iProxies < mappableChildren.Length; iProxies++)
				{
					string classname = mappableChildren[iProxies].GetObjectClassName();

					if (classTable.Contains(classname))
					{
						classTable[classname] =
							new Rational.Test.Ft.Value.RecognitionIndex(
							((Rational.Test.Ft.Value.RecognitionIndex)(classTable[classname])).GetRecognitionIndex() + 1);
					}
					else
					{
						classTable[classname] = new Rational.Test.Ft.Value.RecognitionIndex(0);
					}

					mappableChildren[iProxies].AddRecognitionProperty(CLASSINDEX,
						classTable[classname],
						50);
				}
			}
			return (mappableChildren);
		}


		#endregion

		#region Recording Methods


		public override void ProcessMouseEvent(IMouseActionInfo action)
		{
			currentMethod = null;

			switch (action.GetEventState())
			{
				case MouseActionStates.PRE_DOWN:
					if (action.GetClickCount() == 1)
						ProcessPreDownMouseEvent(action);
					break;
				case MouseActionStates.PRE_UP:
					if (action.GetClickCount() != 1 || action.IsDrag() ||
						initExpandStateSet)
						ProcessPreUpMouseEvent(action);
					break;
				case MouseActionStates.POST_UP:
					if ((this.isColumnResize || this.isRowResize || initParentRowVisibleSet || initDataTableSet) &&
						(!initHitTestInfo.Type.Equals(DataGrid.HitTestType.Cell)))
						ProcessPostUpMouseEvent(action);
					break;
				case MouseActionStates.HOVER:
					ProcessHoverMouseEvent(action);
					break;
			}
		}



		/// <summary>
		/// Process the Pre-down mouse event and determine what action might
		/// be recorded for this object (if any).  The action may be changed
		/// as subsequent events are processed.
		/// </summary>
		/// <param name="action"></param>
		protected override void ProcessPreDownMouseEvent(IMouseActionInfo action)
		{
			IMouseEventInfo event0 = action.GetEventInfo(0);

			int eventCount = action.GetEventCount() - 1;
			IMouseEventInfo eventN = action.GetEventInfo(eventCount);

			// get the first click point
			System.Drawing.Point firstPoint = new System.Drawing.Point(event0.GetX(), event0.GetY());

			GetInitStates(firstPoint);

			// get and save clickArgs on the first PRE_DOWN (may need this for PRE_UP processing)
			clickArgs = GetActionArgs(firstPoint);

			// Can override this method to set more cache variable as needed for derived proxy.
			SetRecordCache(clickArgs, firstPoint);

			// record a click
			currentMethod = "click";

			preDownState = GetScriptCommandFlags();

			SetMethodSpecification(action, currentMethod, SetClickArgs(eventN.GetModifiers()), preDownState);
		}

		/// <summary>
		/// Process the Pre-mouse up event and determine what action should be
		/// recorded for this object (if any).
		/// </summary>
		/// <param name="action">recorder supplied ActionInfo</param>
		protected override void ProcessPreUpMouseEvent(IMouseActionInfo action)
		{
			System.Collections.ArrayList args = null;

			int eventCount = action.GetEventCount() - 1;

			// get first event
			IMouseEventInfo firstEvent = action.GetEventInfo(0);

			// get last event
			IMouseEventInfo lastEvent = action.GetEventInfo(eventCount);

			// get mouse modifiers
			int modifiers = lastEvent.GetModifiers();

			// get first event point
			System.Drawing.Point firstPoint = new System.Drawing.Point(firstEvent.GetX(), firstEvent.GetY());

			// get last event point
			System.Drawing.Point lastPoint = new System.Drawing.Point(lastEvent.GetX(), lastEvent.GetY());

			// multiClickActions - more than two clicks in one place, or a double click + drag
			int clicks = action.GetClickCount();
			if (clicks > 2 || (action.IsDrag() && clicks > 1))
			{
				args = SetMultiClickAction(action, clicks, modifiers, action.IsDrag(), firstPoint, lastPoint);
			}
			else if (clicks == 2 && !action.IsDrag()) // double click
			{
				// should already have this from the first PRE_DOWN
				if (clickArgs == null)
					clickArgs = GetActionArgs(firstPoint);

				args = SetClickArgs(modifiers);
				currentMethod = "doubleClick";
			}
			else if (action.IsDrag() && this.isColumnResize || this.isRowResize)
			{
				clickArgs = GetActionArgs(firstPoint);
				args = SetClickArgs(modifiers);
				currentMethod = "drag";
			}
			else if (action.IsDrag()) // single click drag
			{
				debug.Verbose("Recording drag");
				args = SetDrag(action, modifiers, firstPoint, lastPoint);
			}
			else
			{
				clickArgs = GetActionArgs(firstPoint);
				if (this.isSelectButton)
				{
					args = new System.Collections.ArrayList(20);
					args.Add(Action.Select());
					args.AddRange(clickArgs);
					currentMethod = "setState";
				}
				else if (this.isCollapse || this.isExpand)
				{
					currentMethod = "setState";
					args = new System.Collections.ArrayList(20);

					if (this.isCollapse)
						args.Add(Action.Collapse());
					else
						args.Add(Action.Expand());
					args.AddRange(clickArgs);
				}
				else
				{
					args = SetClickArgs(modifiers);
					currentMethod = "click";
				}
			}

			if (currentMethod != null)
			{
				SetMethodSpecification(action, currentMethod, args, preDownState);
			}
		}

		/// <summary>
		/// Process the Pre-mouse up event and determine what action should be
		/// recorded for this object (if any).
		/// </summary>
		/// <param name="action">recorder supplied ActionInfo</param>
		protected void ProcessPostUpMouseEvent(IMouseActionInfo action)
		{
			debug.Verbose(GetType().FullName + ".ProcessPostUpMouseEvent()");
			int eventCount = action.GetEventCount() - 1;
			System.Collections.ArrayList args = null;

			// get first event
			IMouseEventInfo firstEvent = action.GetEventInfo(0);

			// get last event
			IMouseEventInfo lastEvent = action.GetEventInfo(eventCount);

			// get mouse modifiers
			int modifiers = lastEvent.GetModifiers();

			// get first event point
			System.Drawing.Point firstPoint = new System.Drawing.Point(firstEvent.GetX(), firstEvent.GetY());

			// get last event point
			System.Drawing.Point lastPoint = new System.Drawing.Point(lastEvent.GetX(), lastEvent.GetY());

			int clicks = action.GetClickCount();
			if (action.IsDrag() && this.isColumnResize || this.isRowResize)
			{
				clickArgs = GetActionArgs(firstPoint);
				args = SetClickArgs(modifiers);
				currentMethod = "drag";
			}
			else
			{
				clickArgs = GetActionArgs(firstPoint);
				if (this.isSelectButton)
				{
					args = new System.Collections.ArrayList(20);

					args.Add(Action.Select());
					args.AddRange(clickArgs);
					currentMethod = "setState";
				}
				else
					return;
			}
			if (currentMethod != null)
			{
				SetMethodSpecification(action, currentMethod, args, preDownState);
			}
		}


		protected override System.Collections.ArrayList GetActionArgs(System.Drawing.Point point)
		{
			System.Collections.ArrayList args = new System.Collections.ArrayList();
			Subitem subitem = null;
			if (this.isColumnResize || this.isRowResize)
				subitem = this.GetSubitemForResize(this.isColumnResize);
			else
				subitem = GetSubitemAtPoint(point);

			if (subitem != null)
			{
				args.Add(subitem);

				System.Drawing.Point subitemPoint = System.Drawing.Point.Empty;

				if (subitem is Row ||
					subitem is Rational.Test.Ft.Script.ColumnHeader ||
					isInsideBoolCell)
				{
					subitemPoint = PointToSubitem(subitem, point);
					if (!subitemPoint.IsEmpty)
						args.Add(subitemPoint);
				}
				return args;
			}

			return base.GetActionArgs(point);
		}


		#endregion

		#region Playback Methods

		public override void Drag(MouseModifiers modifiers, Subitem subitem)
		{
			if (subitem is List)
			{
				List list = (List)subitem;
				if (list.GetSubitemCount() != 3)
					base.Drag(modifiers, subitem);
				int pos = 0;
				GetObjectReadyForAction();
				System.Drawing.Point pt1 = System.Drawing.Point.Empty;
				System.Drawing.Point pt2 = System.Drawing.Point.Empty;
				if (list.GetFirstSubitem() is Rational.Test.Ft.Script.ColumnHeader)
				{
					if (list.GetSubitem(1) is Location)
					{
						if (!((Location)list.GetSubitem(1)).IsRightEdge())
							throw new UnsupportedSubitemException(subitem);
					}
					else
						throw new UnsupportedSubitemException(subitem);
					if (list.GetLastSubitem() is Position)
						pos = ((Position)list.GetLastSubitem()).GetPosition();
					SubitemInfo info = FindSubitem(list.GetFirstSubitem());
					if (info == null)
						throw new SubitemNotFoundException(subitem);
					Rectangle rect = this.GetClippedSubitemRect(info);
					pt1 = new System.Drawing.Point(rect.Right, rect.Top + rect.Height / 2);
					pt2 = new System.Drawing.Point(rect.Right + (pos - rect.Width), rect.Top + rect.Height / 2);
					if (pt1.IsEmpty || pt2.IsEmpty)
					{
						throw new UnableToPerformActionException("no point found");
					}
					else
					{
						Rational.Test.Ft.Sys.Graphical.Screen.Get().Drag(modifiers, pt1, pt2);
						return;
					}
				}
				else if (list.GetFirstSubitem() is RowHeader)
				{
					if (list.GetSubitem(1) is Location)
					{
						if (!((Location)list.GetSubitem(1)).IsBottomEdge())
							throw new UnsupportedSubitemException(subitem);
					}
					else
						throw new UnsupportedSubitemException(subitem);
					if (list.GetLastSubitem() is Position)
						pos = ((Position)list.GetLastSubitem()).GetPosition();
					SubitemInfo info = FindSubitem(list.GetFirstSubitem());
					if (info == null)
						throw new SubitemNotFoundException(subitem);
					Rectangle rect = this.GetClippedSubitemRect(info);
					pt1 = new System.Drawing.Point(rect.Left + rect.Width / 2, rect.Bottom - 1);
					pt2 = new System.Drawing.Point(rect.Left + rect.Width / 2, rect.Bottom - 1 + pos - rect.Height);

					if (pt1.IsEmpty || pt2.IsEmpty)
					{
						throw new UnableToPerformActionException("no point found");
					}
					else
					{
						Rational.Test.Ft.Sys.Graphical.Screen.Get().Drag(modifiers, pt1, pt2);
						return;
					}

				}
				else
					throw new UnsupportedSubitemException(subitem);
			}
			else
				base.Drag(modifiers, subitem);
		}


		protected override SubitemInfo FindSubitem(Rational.Test.Ft.Script.Subitem subitem)
		{
			CustomDataGrid netGrid = this.theTestObject as CustomDataGrid;

			if (subitem is Cell)
			{
				return FindSubitemForCell((Rational.Test.Ft.Script.Cell)subitem);
			}
			else if (subitem is Rational.Test.Ft.Script.ColumnHeader)
			{
				return FindSubitemForColumnHeader((Rational.Test.Ft.Script.ColumnHeader)subitem);
			}
			else if (subitem is Rational.Test.Ft.Script.RowHeader)
			{
				return FindSubitemForRowHeader((Rational.Test.Ft.Script.RowHeader)subitem);
			}
			else if (subitem is Location)
			{
				return FindSubitemForLocation((Location)subitem);
			}
			else if (subitem is List)
			{
				return this.FindSubitemForList((List)subitem);
			}
			else if (subitem is Rational.Test.Ft.Script.Row)
			{
				return FindSubitemForRow((Row)subitem);
			}
			else
			{
				throw new UnsupportedSubitemException(subitem);
			}
		}



		#endregion

		#region Verification Point Implementation

		protected const string TESTDATA_GRIDDATA = "griddata";
		protected const string TESTDATA_GRIDDATA_DESCRIPTION = "Grid Data";
		protected const string TESTDATA_SELECTED1 = "selected";
		protected const string TESTDATA_SELECTED_DESCRIPTION1 = "Selected Row";

		public override System.Collections.Hashtable GetTestDataTypes()
		{
			System.Collections.Hashtable types = new HashtableEx();

			types.Add(TESTDATA_GRIDDATA, TESTDATA_GRIDDATA_DESCRIPTION);
			// When DataGrid is displaying a ChildTable (when DataSource is DataSet or DataViewManager )
			// SelectedRow property is returning the value of the parent table selection. 
			// Remove the support for ChildTable "Selected Table Cells" DataVP. 
			if (!IsChildTable())
				types.Add(TESTDATA_SELECTED1, TESTDATA_SELECTED_DESCRIPTION1);

			return types;
		}

		
		public override ITestData GetTestData(string testDataType)
		{
			ITestData data = null;
			switch(testDataType)
			{
				case TESTDATA_GRIDDATA:
					data = GetGridData();
					break;
				case TESTDATA_SELECTED1:
					data = GetSelectedData();
					break;
				default:
					data = base.GetTestData (testDataType);
					break;
			}
			return data;
		}

		
		private ITestDataTable GetGridData()
		{
			TestDataTable data = new TestDataTable("Table", TESTDATA_GRIDDATA_DESCRIPTION, TESTDATA_GRIDDATA);
			SDK.Sample.CustomDataGrid theGrid = this.theTestObject as SDK.Sample.CustomDataGrid;
			Rational.Test.Ft.Domain.Win.IWinControl winGrid = GetWinControlForGrid();
			if (winGrid != null)
			{
				
				int colCount = theGrid.VisibleColumnCount;
				int rowCount = theGrid.VisibleRowCount;

				object[] rowElements;
				rowElements = new object[colCount];

				for (int row = 0; row < rowCount; ++row)
				{
					for (int col = 0; col < colCount; ++col)
					{
						object item = null;
						try
						{
							item = theGrid[row, col];
							if (item != null &&
								!RegisteredConverters.IsRegisteredClassName(item.GetType()))
								item = item.ToString();
						}
						catch (System.Exception)
						{
						}
						rowElements[col] = item;
					}
					data.Add(rowElements);
				}

				for (int col = 0; col < colCount; ++col)
				{
					Rational.Test.Ft.Domain.Win.IWinControl columnHeader = AccessibilityHelper.GetWinColumnHeader(winGrid, col);
					string columnName = string.Empty;
					if (columnHeader != null)
						columnName = columnHeader.Name;
					if (columnName != null && !columnName.Equals(string.Empty))
						data.SetColumnHeader(col, columnName);
				}
			}
			data.AddComparisonRegion(TestDataTableRegion.AllCells());
			data.SetCompareBothByLeftRegions(true);
			return data;
		}

		
		private ITestDataTable GetSelectedData()
		{
			TestDataTable data = new TestDataTable("Table", TESTDATA_SELECTED_DESCRIPTION1, TESTDATA_SELECTED1);
			SDK.Sample.CustomDataGrid theGrid = ((SDK.Sample.CustomDataGrid)this.theTestObject);
			Rational.Test.Ft.Domain.Win.IWinControl winGrid = GetWinControlForGrid();
			if (winGrid != null)
			{
				int colCount = theGrid.VisibleColumnCount;

				int selected = theGrid.CurrentRowIndex;
				object[] rowElements;
				rowElements = new object[colCount];

				for (int col = 0; col < colCount; ++col)
				{
					object item = null;
					try
					{
						item = theGrid[selected, col];
						if (item != null &&
							!RegisteredConverters.IsRegisteredClassName(item.GetType()))
							item = item.ToString();
					}
					catch (System.Exception e)
					{
						debug.Debug(e.StackTrace);
					}
					rowElements[col] = item;
				}
				data.Add(rowElements);

				for (int col = 0; col < colCount; ++col)
				{
					Rational.Test.Ft.Domain.Win.IWinControl columnHeader = AccessibilityHelper.GetWinColumnHeader(winGrid, col);
					string columnName = string.Empty;
					if (columnHeader != null)
						columnName = columnHeader.Name;
					if (columnName != null && !columnName.Equals(string.Empty))
						data.SetColumnHeader(col, columnName);
				}

				Subitem row = GetSubitemForRow(selected);
				if (row != null)
				{
					Row selectedRow = new Row(row);
					data.AddComparisonRegion(TestDataTableRegion.Row(selectedRow));
					if (selectedRow.IsKeyValuePairs())
						data.SetKeyColumns(GetKeyColumns(selected));
				}
			}
			data.SetCompareBothByLeftRegions(true);
			return data;
		}


		
		private bool IsChildTable()
		{
			CustomDataGrid netGrid = this.theTestObject as CustomDataGrid;
			string dataMember = netGrid.DataMember;

			System.Object source = netGrid.DataSource;
			if (source is System.Data.DataSet)
			{
				System.Data.DataRelationCollection relations = ((System.Data.DataSet)source).Relations;
				for (int i = 0; i < relations.Count; i++)
				{
					if (dataMember.EndsWith(relations[i].RelationName))
						return true;
				}
			}
			else if (source is System.Data.DataViewManager)
			{
				System.Data.DataRelationCollection relations = ((System.Data.DataViewManager)source).DataSet.Relations;
				for (int i = 0; i < relations.Count; i++)
				{
					if (dataMember.EndsWith(relations[i].RelationName))
						return true;
				}
			}
			return false;
		}

		#endregion

		#region Data Driving Support

		public void SetValue(int colIndex, int rowIndex, string valueToSet)
		{
			SDK.Sample.CustomDataGrid myTestObject = (SDK.Sample.CustomDataGrid)theTestObject;
			try
			{
				myTestObject[rowIndex,colIndex] = valueToSet;
			}
			catch(Exception ex)
			{
				
			}
		}

		protected override void AddDataDrivableCommands(ProxyTestObject proxy, ArrayList commands, int includeChildren)
		{
			base.AddDataDrivableCommands (proxy, commands, includeChildren);
		}


        
		#endregion

		#region Private Verification Point Methods

		private Subitem[] GetSubitemForRowAndColumn(int rowIndex, int columnIndex)
		{
			if (rowIndex == -1 && columnIndex == -1)
				return null;
			SDK.Sample.CustomDataGrid myTestObject = (SDK.Sample.CustomDataGrid)theTestObject;
			Subitem[] rowColSubitem = new Subitem[2];
			string columnHeaderText = string.Empty;
			string rowHeaderText = string.Empty;
			AccessibleObject gridAccObj = myTestObject.AccessibilityObject;
			int childCount = gridAccObj.GetChildCount();
			int columnCount = 0;
			int rowCount = 0;
			System.Collections.Hashtable columnHeaderHsh = new System.Collections.Hashtable();
			System.Collections.Hashtable rowHeaderHsh = new System.Collections.Hashtable();
			bool bColDone = columnIndex == -1;
			bool bRowDone = rowIndex == -1;

			for (int i = 0; i < childCount; i++)
			{
				AccessibleObject accObjChild = gridAccObj.GetChild(i);

				if (accObjChild.Name != null && accObjChild.Name.Length > 0)
				{
					if (!bColDone && accObjChild.Role == AccessibleRole.ColumnHeader)
					{
						if (columnIndex == columnCount)
						{
							columnHeaderText = accObjChild.Name;
							bColDone = true;
							if (bRowDone)
								break;
						}
						else
						{
							if (!columnHeaderHsh.ContainsKey(accObjChild.Name))
								columnHeaderHsh[accObjChild.Name] = 1;
							else
								columnHeaderHsh[accObjChild.Name] = 1 + (int)columnHeaderHsh[accObjChild.Name];
						}
						columnCount++;
					}
					if (!bRowDone && accObjChild.Role == AccessibleRole.RowHeader)
					{
						if (rowIndex == rowCount)
						{
							rowHeaderText = accObjChild.Name;
							bRowDone = true;
							if (bColDone)
								break;
						}
						else
						{
							if (!rowHeaderHsh.ContainsKey(accObjChild.Name))
								rowHeaderHsh[accObjChild.Name] = 1;
							else
								rowHeaderHsh[accObjChild.Name] = 1 + (int)rowHeaderHsh[accObjChild.Name];
						}
						rowCount++;
					}
				}
			}
			if (columnIndex != -1)
			{
				if (columnHeaderText != null && columnHeaderText.Trim().Length > 0)
				{
					if (columnHeaderHsh.ContainsKey(columnHeaderText))
					{
						rowColSubitem[1] = new Text(columnHeaderText, (int)columnHeaderHsh[columnHeaderText]);
					}
					else
					{
						rowColSubitem[1] = new Text(columnHeaderText);
					}
				}
				else
				{
					rowColSubitem[1] = new Index(columnIndex);
				}
			}

			if (rowIndex != -1)
			{
				if (rowHeaderText != null && rowHeaderText.Trim().Length > 0)
				{
					if (columnHeaderHsh.ContainsKey(rowHeaderText))
					{
						rowColSubitem[0] = new Text(columnHeaderText, (int)columnHeaderHsh[rowHeaderText]);
					}
					else
					{
						rowColSubitem[0] = new Text(rowHeaderText);
					}
				}
				else
				{
					rowColSubitem[0] = GetSubitemForRowWithoutHeader(rowIndex);
				}
			}
			return rowColSubitem;
		}

		private Subitem GetSubitemForRowWithoutHeader(int rowIndex)
		{
			int[] keyCols = GetKeyColumns(rowIndex);
			System.Data.DataTable table = GetCurrentDataTable();
			if (keyCols != null && table != null)
			{
				int cols = keyCols.Length;
				string[] colValues = new string[cols];
				System.Collections.ArrayList keyValue = new System.Collections.ArrayList(20);
				string name;
				for (int i = 0; i < keyCols.Length; i++)
				{
					name = table.Columns[keyCols[i]].ColumnName;
					if (name != null)
					{
						string item = table.Rows[rowIndex][keyCols[i]].ToString();
						if (!item.Equals(String.Empty) && item.Length > 0)
						{
							keyValue.Add(name);
							keyValue.Add(item);
							colValues[i] = item;
						}
					}
				}

				int size = keyValue.Count;
				object[] keyValuePairs = new object[size];
				for (int i = 0; i < size; i++)
					keyValuePairs[i] = keyValue[i];
				if (TestKeysForUniqueness(rowIndex, keyCols.Length, keyCols, colValues))
					return new Row(keyValuePairs);
			}
			return new Index(rowIndex);
		}


		private System.Data.DataTable GetCurrentDataTable()
		{
			CustomDataGrid netGrid = this.theTestObject as CustomDataGrid;
			string dataMember = netGrid.DataMember;

			System.Object source = netGrid.DataSource;
			if (source is System.Data.DataTable)
			{
				return (System.Data.DataTable)source;
			}
			else if (source is System.Data.DataSet)
			{
				System.Data.DataTableCollection tables = ((System.Data.DataSet)source).Tables;

				for (int i = 0; i < tables.Count; i++)
				{
					if (tables[i].TableName.Equals(dataMember))
						return tables[i];
				}
				System.Data.DataRelationCollection relations = ((System.Data.DataSet)source).Relations;
				for (int i = 0; i < relations.Count; i++)
				{
					if (dataMember.EndsWith(relations[i].RelationName))
						return relations[i].ChildTable;
				}
			}
			else if (source is System.Data.DataView)
			{
				return ((System.Data.DataView)source).Table;
			}
			else if (source is System.Data.DataViewManager)
			{
				System.Data.DataTableCollection tables = ((System.Data.DataViewManager)source).DataSet.Tables;
				for (int i = 0; i < tables.Count; i++)
				{
					if (tables[i].TableName.Equals(dataMember))
						return tables[i];
				}
				System.Data.DataRelationCollection relations = ((System.Data.DataViewManager)source).DataSet.Relations;
				for (int i = 0; i < relations.Count; i++)
				{
					if (dataMember.EndsWith(relations[i].RelationName))
						return relations[i].ChildTable;
				}
			}
			return null;
		}

		private Subitem GetSubitemForColumn(int columnIndex)
		{
			Subitem[] rowColSubitem = GetSubitemForRowAndColumn(-1, columnIndex);
			return rowColSubitem[1];
		}

		private Subitem GetSubitemForRow(int rowIndex)
		{
			Subitem[] rowColSubitem = GetSubitemForRowAndColumn(rowIndex, -1);
			return rowColSubitem[0];
		}


		private Rational.Test.Ft.Domain.Win.IWinControl GetWinControlForGrid()
		{
			Rational.Test.Ft.Domain.Win.IWinControl winGrid = null;
			SDK.Sample.CustomDataGrid myTestObject = (SDK.Sample.CustomDataGrid)theTestObject;
			Rational.Test.Ft.Domain.Win.IWinControl obj = AccessibilityHelper.GetAccObj(myTestObject.Handle);
			if (obj != null)
			{
				Rational.Test.Ft.Domain.Win.IWinControl[] children = AccessibilityHelper.GetAccChildren(obj);
				if (children != null && children.Length > 0)
				{
					for (int i = 0; i < children.Length; i++)
					{
						if (children[i].AccessibleRole.Equals(".Table"))
						{
							winGrid = children[i];
							break;
						}
					}
				}
			}
			return winGrid;
		}
		

		private int[] GetKeyColumns(int nRow)
		{
			System.Data.DataTable table = GetCurrentDataTable();
			if (table == null)
				return null;
			if (table.Rows.Count <= nRow)
				return null;
			int[] roCols = new int[3];
			int roCol = 0;

			int nCols = table.Columns.Count;
			for (int nIndex = 0; nIndex < nCols; ++nIndex)
			{
				if (table.Columns[nIndex].Unique)
				{
					roCols[roCol] = nIndex;
					object item = table.Rows[nRow][nIndex];
					string text = item.ToString();

					if (text != null && !text.Equals(String.Empty))
						++roCol;
				}
				if (roCol == 3) break; // once we have three KEYS quit looking
			}

			if (roCol != 0)
			{
				int[] nResult = new int[roCol];
				for (int nIndex = 0; nIndex < roCol; ++nIndex)
					nResult[nIndex] = roCols[nIndex];

				return nResult;
			}
			else
				return null;
		}

		
		private bool TestKeysForUniqueness(int row, int roCol, int[] roCols,
			string[] roValues)
		{
			// Make sure that the keys are unique up to the row that was
			// selected.
			bool unique = false;
			System.Data.DataTable table = GetCurrentDataTable();
			for (int col = 0; col < roCol; col++)
			{
				if (table.Columns[roCols[col]].Unique)
					// when one of the key columns is unique, then the pairs is unique
				{
					unique = true;
					break;
				}
			}
			if (unique) return true;

			string key1Value = roValues[0];
			string key2Value = null;
			if (roCol >= 2) key2Value = roValues[1];
			string key3Value = null;
			if (roCol >= 3) key3Value = roValues[2];

			for (int nRow = 0; nRow < row; ++nRow)
			{
				string item = table.Rows[nRow][roCols[0]].ToString();
				if (key1Value.Equals(item))
				{
					if (key2Value != null)
					{
						item = table.Rows[nRow][roCols[1]].ToString();
						if (key2Value.Equals(item.ToString()))
						{
							if (key3Value != null)
							{
								item = table.Rows[nRow][roCols[2]].ToString();
								if (key3Value.Equals(item.ToString()))
								{
									return false; // three keys not unique
								}
							}
							else  // third key missing and first two not unique
								return false;
						}
					}
					else // no second key and first key not unique
						return false;
				}
			}
			return true;
		}

		#endregion

		#region Private Recording Methods

		private DataGrid.HitTestInfo initHitTestInfo = null;

		private bool initExpandState = false;
		private bool initExpandStateSet = false;
		private string initDataTable = String.Empty;
		private bool initDataTableSet = false;
		private bool initParentRowVisible = false;
		private bool initParentRowVisibleSet = false;
		private bool isExpand = false;
		private bool isCollapse = false;
		private bool isInsideBoolCell = false;
		private bool isSelectButton = false;
		private bool isColumnResize = false;
		private bool isRowResize = false;

		private bool scrollSearchFinished = false;
		private CScrollBar vScrollBar = null;
		private CScrollBar hScrollBar = null;
		bool hScrollFound = false;
		bool vScrollFound = false;

		private void InitializeMembers()
		{
			this.initHitTestInfo = null;
			this.initExpandState = false;
			this.initExpandStateSet = false;
			this.initDataTable = String.Empty;
			this.initDataTableSet = false;
			this.initParentRowVisible = false;
			this.initParentRowVisibleSet = false;
			this.isExpand = false;
			this.isCollapse = false;
			this.isInsideBoolCell = false;
			this.isSelectButton = false;
			this.isColumnResize = false;
			this.isRowResize = false;

			this.scrollSearchFinished = false;
			this.vScrollBar = null;
			this.hScrollBar = null;
			this.hScrollFound = false;
			this.vScrollFound = false;
		}


		private void GetInitStates(System.Drawing.Point point)
		{
			CustomDataGrid netGrid = this.theTestObject as CustomDataGrid;
			InitializeMembers();
			initHitTestInfo = netGrid.HitTest(netGrid.PointToClient(point));


			if (initHitTestInfo.Type.Equals(DataGrid.HitTestType.Cell))
			{
				System.Data.DataTable theTable = GetCurrentDataTable();
				if (theTable != null)
				{
					System.Type type = theTable.Columns[initHitTestInfo.Column].DataType;
					if (type.Name.EndsWith("Boolean"))
					{
						isInsideBoolCell = true;
					}
				}
			}
			else if (netGrid.DataSource is System.Data.DataSet ||
				netGrid.DataSource is System.Data.DataViewManager ||
				netGrid.DataSource is System.Data.DataView)
			{
				if (initHitTestInfo.Type.Equals(DataGrid.HitTestType.Caption))
				{
					initDataTable = netGrid.DataMember;
					initDataTableSet = true;
					initParentRowVisible = netGrid.ParentRowsVisible;
					initParentRowVisibleSet = true;
				}
				else if (initHitTestInfo.Type.Equals(DataGrid.HitTestType.RowHeader))
				{
					initExpandState = netGrid.IsExpanded(initHitTestInfo.Row);
					initExpandStateSet = true;
				}
			}
		}


		private Subitem GetSubitemForResize(bool isColumnResize)
		{
			if (this.initHitTestInfo == null)
				return null;
			if (isColumnResize)
			{
				Subitem item = GetSubitemForColumn(initHitTestInfo.Column);
				if (GetWinControlForGrid() == null)
					return null;
				Rational.Test.Ft.Domain.Win.IWinControl header = AccessibilityHelper.GetWinColumnHeader(GetWinControlForGrid(), initHitTestInfo.Column);
				if (item == null || header == null)
					return null;
				else
					return new List(new Rational.Test.Ft.Script.ColumnHeader(item),
						Location.GetRightEdge(),
						new Rational.Test.Ft.Script.Position(header.Rectangle.Width));
			}
			else
			{
				Subitem item = GetSubitemForRow(initHitTestInfo.Row);
				AccessibleObject row = AccessibilityHelper.GetAccessibleObjectForRow((Control)this.theTestObject, initHitTestInfo.Row);
				if (item == null || row == null)
					return null;
				else
					return new List(new Rational.Test.Ft.Script.RowHeader(item),
						Location.GetBottomEdge(),
						new Rational.Test.Ft.Script.Position(row.Bounds.Height));
			}
		}

		private Subitem GetSubitemInsideRowHeader(DataGrid.HitTestInfo info)
		{
			CustomDataGrid netGrid = this.theTestObject as CustomDataGrid;
			

			if (netGrid.IsExpanded(info.Row) != this.initExpandState)
			{
				if (initExpandState)
					isCollapse = true;
				else
					isExpand = true;
				return new Row(GetSubitemForRow(info.Row));
			}
			Subitem item = GetSubitemForRow(info.Row);
			return (item == null ? item : (new Rational.Test.Ft.Script.RowHeader(item)));
		}

		private Subitem GetSubitemInsideCaption(DataGrid.HitTestInfo info)
		{
			DataGrid netGrid = (DataGrid)this.theTestObject;

			Location loc = new Location(Location.CAPTION);
			if (this.initDataTableSet)
			{
				string currentDataTable = netGrid.DataMember;
				if (!initDataTable.Equals(currentDataTable))
				{
					this.isSelectButton = true;
					return Location.GetBackButton();
				}
				else if (netGrid.ParentRowsVisible != this.initParentRowVisible)
				{
					this.isSelectButton = true;
					return Location.GetShowHideButton();
				}
			}
			return loc;
		}

		private Subitem GetSubitemInsideCell(System.Drawing.Point point, DataGrid.HitTestInfo info)
		{
            
			Cell cellSubitem = GetSubitemForCell(info.Column, info.Row);
			if (cellSubitem == null)
				return null;
			CustomDataGrid myTestObject = (CustomDataGrid)theTestObject;
			AccessibleObject accObjChild = null;
			Rectangle accObjChildRect = Rectangle.Empty;
			AccessibleObject accObjRow = AccessibilityHelper.GetAccessibleObjectForRow((Control)this.theTestObject,info.Row);
			if (FtDebug.DEBUG)
			{
				debug.Debug(string.Format("GetSubitemInsideCell called with point={0},row={1},column={2}", point, info.Row, info.Column));
			}
			for (int i = 0; i < accObjRow.GetChildCount(); i++)
			{
				accObjChild = accObjRow.GetChild(i);
				if (FtDebug.DEBUG)
				{
					debug.Debug(string.Format("Child at index {0},Name ={1},Role={2},Bounds ={3},Value ={4}", i, accObjChild.Name, accObjChild.Role, accObjChild.Bounds,accObjChild.Value));
				}
				if (accObjChild.Role == AccessibleRole.Link)
				{
					accObjChildRect = myTestObject.RectangleToClient(accObjChild.Bounds);
					if (FtDebug.DEBUG)
					{
						debug.Debug(string.Format("Child at index {0},Name ={1},Role={2},Bounds ={3},ClientRect={4}", i, accObjChild.Name, accObjChild.Role, accObjChild.Bounds, accObjChildRect));
					}
					if (accObjChild.Bounds.Contains(point))
					{
						Subitem retSubitem = new List(cellSubitem.GetRow(), new Text(accObjChild.Name));
						if (FtDebug.DEBUG)
						{
							debug.Debug("Return From LINK ROLE GetSubitemInsideCell >" + retSubitem.ToString());
						}
						return retSubitem;
					}
				}
				else if (accObjChild.Role == AccessibleRole.Cell)
				{
					accObjChildRect = myTestObject.RectangleToClient(accObjChild.Bounds);
					if (accObjChild.Bounds.Contains(point))
					{
						if (FtDebug.DEBUG)
						{
							debug.Debug("Return From CELL ROLE GetSubitemInsideCell >" + cellSubitem.ToString());
						}
						return cellSubitem;
					}
				}
			}
			if (FtDebug.DEBUG)
			{
				debug.Debug("Return From ROW GetSubitemInsideCell >");
			}

			Subitem row = GetSubitemForRow(info.Row);
			return (row == null ? null : (new Row(row)));
		}

		private Rational.Test.Ft.Script.Cell GetSubitemForCell(int columnIndex, int rowIndex)
		{
			Subitem[] rowColSubitem = GetSubitemForRowAndColumn(rowIndex, columnIndex);
			Subitem col = rowColSubitem[1];//GetSubitemForColumn(columnIndex);
			Subitem row = rowColSubitem[0];//GetSubitemForRow( rowIndex );
			if (col == null || row == null)
				return null;
			return new Rational.Test.Ft.Script.Cell(new Column(col), new Row(row));
		}

		private Rational.Test.Ft.Script.Subitem GetSubitemAtPoint(System.Drawing.Point point)
		{
			CustomDataGrid netGrid = this.theTestObject as CustomDataGrid;
			if (!scrollSearchFinished)
				FindChildScrollBars();

			Subitem subitem = null;

			DataGrid.HitTestInfo info = netGrid.HitTest(netGrid.PointToClient(point));

			//debug.Debug("hittestinfo type: " + info.Type );

			switch (info.Type)
			{
				case DataGrid.HitTestType.None:
					subitem = GetScrollbarSubitem(point);
					break;
				case DataGrid.HitTestType.Cell:
					subitem = GetSubitemInsideCell(point,info);
					break;
				case DataGrid.HitTestType.ColumnHeader:
					Subitem header = GetSubitemForRowAndColumn(-1, info.Column)[1];
					if (header != null)
						subitem = new Rational.Test.Ft.Script.ColumnHeader(header);
					break;
				case DataGrid.HitTestType.Caption:
					subitem = GetSubitemInsideCaption(info);
					break;
				case DataGrid.HitTestType.ParentRows:
					subitem = Location.GetParentRows();
					break;
				case DataGrid.HitTestType.RowHeader:
					subitem = GetSubitemInsideRowHeader(info);
					break;
				case DataGrid.HitTestType.ColumnResize:
					this.isColumnResize = true;
					this.initHitTestInfo = info;
					break;
				case DataGrid.HitTestType.RowResize:
					this.isRowResize = true;
					this.initHitTestInfo = info;
					break;
			}

			return subitem;
		}

		private System.Drawing.Point PointToSubitem(Subitem subitem, System.Drawing.Point screenPoint)
		{
			DataGrid theGrid = ((DataGrid)this.theTestObject);
			System.Drawing.Point pointToSubitem = System.Drawing.Point.Empty;

			if (subitem is Location)
			{
				if (((Location)subitem).IsCaption())
				{
					pointToSubitem = theGrid.PointToClient(screenPoint);
				}
				else if (((Location)subitem).IsParentRow() && GetWinControlForGrid() != null)
				{
					Rational.Test.Ft.Domain.Win.IWinControl parentRow = AccessibilityHelper.GetWinParentRows(GetWinControlForGrid());
					pointToSubitem.X = screenPoint.X - parentRow.Rectangle.X;
					pointToSubitem.Y = screenPoint.Y - parentRow.Rectangle.Y;
				}
			}
			if (GetWinControlForGrid() != null)
			{
				if (subitem is Header)
				{
					Subitem header = ((Header)subitem).GetHeader();
					if (header is Index) //RowHeader
					{
						int index = ((Index)header).GetIndex();
						AccessibleObject row = AccessibilityHelper.GetAccessibleObjectForRow((Control)this.theTestObject, index);
						Rectangle clientAccObjRect = theGrid.RectangleToClient(row.Bounds);
						pointToSubitem.X = theGrid.PointToClient(screenPoint).X;
						pointToSubitem.Y = screenPoint.Y - clientAccObjRect.Y;
					}
				}
				if (subitem is Row)
				{
					if (this.isCollapse || this.isExpand)
						return pointToSubitem;
					int rowIndex = -1;
					if (((Row)subitem).IsIndex())
						rowIndex = ((Row)subitem).GetIndex().GetIndex();
					if (rowIndex >= 0)
					{
						AccessibleObject row = AccessibilityHelper.GetAccessibleObjectForRow(theGrid,rowIndex);
						Rectangle clientAccObjRect = theGrid.RectangleToClient(row.Bounds);
						pointToSubitem.Y = screenPoint.Y - clientAccObjRect.Y;
						pointToSubitem.X = screenPoint.X - clientAccObjRect.X;
					}
				}
				if (subitem is Cell)
				{
					Rectangle cell = theGrid.GetCellBounds(this.initHitTestInfo.Row, this.initHitTestInfo.Column);
					pointToSubitem.X = theGrid.PointToClient(screenPoint).X - cell.X;
					pointToSubitem.Y = theGrid.PointToClient(screenPoint).Y - cell.Y;
				}
			}
			return pointToSubitem;
		}

		private void FindChildScrollBars()
		{
			CustomDataGrid netGrid = this.theTestObject as CustomDataGrid;
			System.Windows.Forms.Control.ControlCollection children = netGrid.Controls;

			for (int i = 0; i < children.Count; i++)
			{
				if (children[i] is System.Windows.Forms.VScrollBar)
				{
					vScrollBar = new CScrollBar(children[i], true);
					vScrollFound = true;
				}
				if (children[i] is System.Windows.Forms.HScrollBar)
				{
					hScrollBar = new CScrollBar(children[i], false);
					hScrollFound = true;
				}
				if (vScrollFound && hScrollFound)
					break;
			}
			scrollSearchFinished = true;
		}



		#endregion

		#region Private Playback Methods

		private void FindCellIndex(Cell cell, ref int columnIndex, ref int rowIndex)
		{
			columnIndex = -1;
			rowIndex = -1;
			Column col = cell.GetColumn();
			Row row = cell.GetRow();

			columnIndex = FindColumnIndex(col);
			rowIndex = FindRowIndex(row);
			//debug.Debug("subitem is: " + cell + "column index is: " + columnIndex + "rowindex is: " + rowIndex);
		}

		private CellSubitemInfo FindSubitemForCell(Rational.Test.Ft.Script.Cell cell)
		{
			CellSubitemInfo info = new CellSubitemInfo();
			int columnIndex = -1;
			int rowIndex = -1;
			FindCellIndex(cell, ref columnIndex, ref rowIndex);
			if (columnIndex == -1 || rowIndex == -1)
				return null;
			info.SetIndices(columnIndex, rowIndex);
			return info;
		}

		private LinkSubitemInfo FindSubitemForList(Rational.Test.Ft.Script.List list)
		{
			if(FtDebug.DEBUG)
			{
				debug.Debug("FindSubitemForList called for "+ list);
			}
			if (list.GetSubitemCount() == 2 &&
				list.GetFirstSubitem() is Row &&
				list.GetLastSubitem() is Text)
			{
				Row row = (Row)list.GetFirstSubitem();
				Text link = (Text)list.GetLastSubitem();

				int rowIndex = FindRowIndex(row);
				if(FtDebug.DEBUG)
				{
					debug.Debug("Index returned for "+ row +" = "+rowIndex);
				}
				CustomDataGrid myTestObject = theTestObject as CustomDataGrid;

				AccessibleObject accObjRow = AccessibilityHelper.GetAccessibleObjectForRow(myTestObject, rowIndex);

				if (accObjRow != null)
				{
					AccessibleObject accObjChild = null;
					if (accObjRow.GetChildCount() > 0)
					{
						int childCount = accObjRow.GetChildCount();
						for (int i = 0; i < childCount; i++)
						{
							accObjChild = accObjRow.GetChild(i);
							if (FtDebug.DEBUG)
							{
								debug.Debug(string.Format("Child at index {0},Name ={1},Role={2},Bounds ={3},Value ={4}", i, accObjChild.Name, accObjChild.Role, accObjChild.Bounds,accObjChild.Value));
							}
							if (accObjChild.Role == AccessibleRole.Link)
							{
								if (accObjChild.Name.Equals(link.GetText()))
								{
									LinkSubitemInfo info = new LinkSubitemInfo();
									//info.SetWinControl(children[i]);
									info.SetSubitemRect(myTestObject.RectangleToClient(accObjChild.Bounds));// Nitin Check it out
									if (FtDebug.DEBUG)
									{
										debug.Debug(string.Format("Return from Link {0}",info.GetSubitemRect()));
									}
									return info;
								}
							}
						}
					}
				}
				if (FtDebug.DEBUG)
				{
					debug.Debug(string.Format("Return NULL"));
				}              
				return null;
			}
			else
				throw new UnsupportedSubitemException(list);
		}

		private ColumnHeaderSubitemInfo FindSubitemForColumnHeader(Rational.Test.Ft.Script.ColumnHeader header)
		{
			ColumnHeaderSubitemInfo info = new ColumnHeaderSubitemInfo();
			CustomDataGrid myTestObject = theTestObject as CustomDataGrid;
			AccessibleObject accObjGrid = myTestObject.AccessibilityObject;
			int childCount = accObjGrid.GetChildCount();
			AccessibleObject accObjChild = null;

			int counter = -1;

			if (header.GetHeader() is Index)
			{
                
				int index = ((Index)header.GetHeader()).GetIndex();
				for (int i = 0; i < childCount; i++)
				{
					accObjChild = accObjGrid.GetChild(i);
					if (accObjChild.Role == AccessibleRole.ColumnHeader)
					{
						counter++;
						if (counter == index)
						{
							info.SetColumnIndex(index);
							info.SetSubitemRect(myTestObject.RectangleToClient(accObjChild.Bounds));
							return info;
						}
					}
				}
			}
			else if (header.GetHeader() is Text)
			{
				int textIndex = ((Text)header.GetHeader()).GetIndex();
				string text = ((Text)header.GetHeader()).GetText();
				for (int i = 0; i < childCount; i++)
				{
					accObjChild = accObjGrid.GetChild(i);
					if (accObjChild.Role == AccessibleRole.ColumnHeader)
					{
						if(accObjChild.Name.Equals(text))
						{
							counter++;
							if (textIndex == 0)
							{
								info.SetColumnIndex(counter);
								info.SetSubitemRect(myTestObject.RectangleToClient(accObjChild.Bounds)); 
								return info;
							}
							textIndex--;
						}
					}
				}
			}
			return info;
		}

		private int FindRowIndex(Subitem subitem)
		{
			int index = -1;
			if (subitem is Rational.Test.Ft.Script.RowHeader)
			{
				return FindRowIndex(((Rational.Test.Ft.Script.RowHeader)subitem).GetHeader());
			}
			else if (subitem is Row)
			{
				if (((Row)subitem).IsSubitem())
					return FindRowIndex(((Rational.Test.Ft.Script.Row)subitem).GetSubitem());
				else if (((Row)subitem).IsIndex())
					return FindRowIndex(((Row)subitem).GetIndex());
				else if (((Row)subitem).IsKeyValuePairs())
				{
					System.Data.DataTable table = GetCurrentDataTable();
					if (table == null)
						throw new SubitemNotFoundException(subitem);
					object[] keyValuePairs = ((Row)subitem).GetKeyValuePairs();
					int roCols = keyValuePairs.Length / 2;
					if (roCols * 2 != keyValuePairs.Length)
						throw new SubitemNotFoundException(subitem);
					object[] keys = new object[roCols];
					object[] values = new object[roCols];
					for (int i = 0; i < roCols; i++)
					{
						keys[i] = keyValuePairs[2 * i];
						values[i] = keyValuePairs[2 * i + 1];
					}
					int[] cols = new int[roCols];
					int counter = 0;
					for (int i = 0; i < table.Columns.Count; i++)
					{
						for (int j = 0; j < roCols; j++)
						{
							if (table.Columns[i].ColumnName.Equals(keys[j]))
							{
								cols[j] = i;
								counter++;
							}
						}
					}
					if (counter < roCols)
						throw new SubitemNotFoundException(subitem);
					for (int i = 0; i < table.Rows.Count; i++)
					{
						int find = 0;
						for (int j = 0; j < roCols; j++)
						{
							if (!table.Rows[i][cols[j]].ToString().Equals(values[j]))
								break;
							else
								find++;
						}
						if (find == roCols)
						{
							index = i;
							break;
						}
					}
					return index;
				}
			}
			else if (subitem is Index)
			{
				CustomDataGrid myTestObject = theTestObject as CustomDataGrid;
				int rowIndex = ((Index)subitem).GetIndex();
				if (myTestObject.VisibleRowCount < rowIndex)
					return -1;
				AccessibleObject accObjGrid = myTestObject.AccessibilityObject;
				int childCount = accObjGrid.GetChildCount();
				index = rowIndex;
				int newRowIndex = -1;
				AccessibleObject accObjChild = null;
				for (int i = 0; i < childCount; i++)
				{
					accObjChild = accObjGrid.GetChild(i);
					if (accObjChild.Role == AccessibleRole.Row)
					{
						if (accObjChild.Name.Equals("(Create New)"))
						{
							newRowIndex = i;
						}
						else
						{
							if (rowIndex == 0)
								return index;
							rowIndex--;
						}
					}
				}
				if (rowIndex == 0 && newRowIndex != -1)
					return index;
			}
			return -1;

		}

		private int FindColumnIndex(Subitem subitem)
		{
			int index = -1;
			if (subitem is Rational.Test.Ft.Script.ColumnHeader)
			{
				return FindColumnIndex(((Rational.Test.Ft.Script.ColumnHeader)subitem).GetHeader());
			}
			else if (subitem is Column)
			{
				if (((Column)subitem).IsSubitem())
					return FindColumnIndex(((Rational.Test.Ft.Script.Column)subitem).GetSubitem());
				else if (((Column)subitem).IsIndex())
					return FindColumnIndex(((Column)subitem).GetIndex());
			}
			else if (subitem is Index)
			{
				int columnIndex = ((Index)subitem).GetIndex();
				CustomDataGrid myTestObject = theTestObject as CustomDataGrid;
				if (columnIndex < myTestObject.VisibleColumnCount)
				{
					index = columnIndex;
				}
			}
			else if (subitem is Text)
			{
				string text = ((Text)subitem).GetText();

				int textIndex = ((Text)subitem).GetIndex();
				if (textIndex < 0)
					throw new InvalidSubitemException(subitem.ToString());
				CustomDataGrid myTestObject = theTestObject as CustomDataGrid;
				AccessibleObject accObjGrid = myTestObject.AccessibilityObject;
				int childCount = accObjGrid.GetChildCount();
				AccessibleObject accObjChild = null;
				for (int i = 0; i < childCount; i++)
				{
					accObjChild = accObjGrid.GetChild(i);
					if (accObjChild.Role == AccessibleRole.ColumnHeader)
					{
						index++;
						if (accObjChild.Name != null && accObjChild.Name.Equals(text))
						{
							if (textIndex == 0)
								break;
							textIndex--;
						}
					}
				}
			}
			return index;
		}


		private RowHeaderSubitemInfo FindSubitemForRowHeader(Rational.Test.Ft.Script.RowHeader header)
		{
			
			RowHeaderSubitemInfo info = new RowHeaderSubitemInfo();
			CustomDataGrid myTestObject = theTestObject as CustomDataGrid;
			AccessibleObject accObjGrid = myTestObject.AccessibilityObject;
			int childCount = accObjGrid.GetChildCount();
			AccessibleObject accObjChild = null;
			int index = this.FindRowIndex(header);
			int counter = -1;
			int newRowIndex = -1;

			for (int i = 0; i < childCount; i++)
			{
				accObjChild = accObjGrid.GetChild(i);
				if (accObjChild.Role== AccessibleRole.Row)
				{
					if (accObjChild.Name.Equals("(Create New)"))
					{
						newRowIndex = i;
					}
					else
					{
						counter++;
						if (counter == index)
						{
							info.SetRowIndex(index);
							//info.SetRowWinControl(children[i]);
							System.Drawing.Rectangle itemRect = accObjChild.Bounds;
							itemRect.Width = myTestObject.RowHeaderWidth;
							itemRect = myTestObject.RectangleToClient(itemRect);
							itemRect.X = 0;
							info.SetSubitemRect(itemRect);
							break;
						}
					}
				}
			}
			if (counter == index - 1 && newRowIndex != -1) // this is "creat new" row 
			{
				info.SetRowIndex(index);
				//info.SetRowWinControl(children[newRowIndex]);
				System.Drawing.Rectangle itemRect = accObjGrid.GetChild(newRowIndex).Bounds;
				itemRect.Width = myTestObject.RowHeaderWidth;
				itemRect = myTestObject.RectangleToClient(itemRect);
				itemRect.X = 0;
				info.SetSubitemRect(itemRect);
			}
			return info;
		}

		private RowSubitemInfo FindSubitemForRow(Rational.Test.Ft.Script.Row row)
		{
			CustomDataGrid myTestObject = theTestObject as CustomDataGrid;
			int index = FindRowIndex(row);
			if (index < 0)
				return null;

			RowSubitemInfo info = new RowSubitemInfo();
			info.SetRowIndex(index);
			AccessibleObject winRow = AccessibilityHelper.GetAccessibleObjectForRow(myTestObject, index);
			if (winRow != null)
			{
				//info.SetWinControl(winRow);
				info.SetSubitemRect(myTestObject.RectangleToClient(winRow.Bounds));
			}
			return info;
		}

		private SubitemInfo FindSubitemForLocation(Rational.Test.Ft.Script.Location loc)
		{
			CustomDataGrid netGrid = theTestObject as CustomDataGrid;
			Rational.Test.Ft.Domain.Win.IWinControl winGrid = GetWinControlForGrid();
			if (winGrid != null)
			{
				LocationSubitemInfo info = new LocationSubitemInfo();

				if (loc.IsScrollLocation())
				{
					if (!scrollSearchFinished)
					{
						FindChildScrollBars();
					}
					if (!(this.vScrollFound || this.hScrollFound))
						return null;

					bool isVertical = (loc.IsScrollDown() || loc.IsScrollDownButton() ||
						loc.IsScrollUp() || loc.IsScrollUpButton() || loc.IsVerticalScrollElevator());

					if ((isVertical && vScrollBar != null && vScrollBar.Visible) ||
						(!isVertical && hScrollBar != null && hScrollBar.Visible))
					{
						ClientScrollSubitemInfo scrollInfo = new ClientScrollSubitemInfo(loc.GetLocation());
						scrollInfo.SetVertical(isVertical);
						return scrollInfo;
					}
					return null;
				}
				if (loc.IsParentRow())
				{
					Rational.Test.Ft.Domain.Win.IWinControl parentRow = AccessibilityHelper.GetWinParentRows(winGrid);
					if (parentRow != null)
					{
						info.SetWinControl(parentRow);
					}
					else
						return null;
					return info;
				}
				else if (loc.IsCaption())
				{
					info.SetCaption(true);
					if (netGrid.CaptionVisible)
					{
						Rectangle rect = netGrid.ClientRectangle;
						if (netGrid.ParentRowsVisible)
						{
							Rational.Test.Ft.Domain.Win.IWinControl parentRow = AccessibilityHelper.GetWinParentRows(GetWinControlForGrid());
							Rectangle parentRowRect = parentRow.Rectangle;
							if (parentRowRect.Height > 0)
							{
								rect.Height = netGrid.RectangleToClient(parentRowRect).Y;
								info.SetSubitemRect(rect);
								return info;
							}
						}
						if (netGrid.ColumnHeadersVisible)
						{
							Rational.Test.Ft.Domain.Win.IWinControl columnHeader = AccessibilityHelper.GetWinColumnHeader(this.GetWinControlForGrid(), 0);
							Rectangle columnHeaderRect = columnHeader.Rectangle;
							if (columnHeaderRect.Height > 0)
							{
								rect.Height = netGrid.RectangleToClient(columnHeaderRect).Y;
								info.SetSubitemRect(rect);
								return info;
							}
						}
						if (!netGrid.GetCellBounds(0, 0).Equals(null))
						{
							rect.Height = netGrid.GetCellBounds(0, 0).Y - rect.Y;
							info.SetSubitemRect(rect);
							return info;
						}
					}
					return null;
				}
				else
					throw new UnsupportedSubitemException(loc);
			}
			return null;
		}


		#endregion

	}
}
